@@ -51,6 +51,13 @@ sub new {
         $vhost = $ENV{MQVHOST} if exists $ENV{MQVHOST};
         $port = $ENV{MQPORT} || 5672;
     }
+    my $admin_protocol = $ENV{MQADMINPROTOCOL} || "https";
+    my $admin_port = $ENV{MQADMINPORT} || "443";
+
+    #hack but it's ok as it's for testing and I don't want more deps
+    my $uri_encoded_vhost = $vhost;
+    $uri_encoded_vhost =~ s|/|%2F|g;
+    my $admin_api_url = "$admin_protocol://$username:$password\@$host:$admin_port/api/exchanges/$uri_encoded_vhost";
 
     my $self = {
         unique             => $unique,
@@ -72,6 +79,7 @@ sub new {
         vhost              => $vhost,
         declared_exchanges => [],
         declared_queues    => [],
+        admin_api_url      => $admin_api_url,
         %options,
     };
     if ( $ENV{NARDEBUG} ) {
